// 003 u_c_xml_tree_nested
// 15 oct 2006

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_xml_tree_nested;
  interface
    uses Classes, u_c_basic_object;

    type c_xml_tag_content_list= Class; // Forward
         c_xml_text= Class; // forward

         // -- a simple string
         c_xml_string= Class(c_basic_object)
                         // -- m_name: the string
                         m_c_parent_xml_text: c_xml_text;

                         Constructor create_xml_string(p_name: String;
                             p_c_parent_xml_text: c_xml_text);
                         function f_c_self: c_xml_string;
                         function f_display_xml_string: String;
                       end; // c_xml_string

         // -- the tag: attributes and content list
         c_xml_tag= Class(c_xml_string)
                      // -- m_name: the tag name

                      // -- the key="value" lists
                      m_c_key_list, m_c_value_list: tStringList;

                      // -- the (optional) content between <xxx> and </xxx>
                      m_c_xml_tag_content_list: c_xml_tag_content_list;

                      // -- XSD INCLUDE
                      m_c_xml_nested_text: c_xml_text;
                      m_is_ref, m_is_type: Boolean;
                      m_c_type_tag, m_c_ref_tag: c_xml_tag;

                      Constructor create_xml_tag(p_name: String;
                          p_c_parent_xml_text: c_xml_text);
                      function f_c_self: c_xml_tag;

                      procedure add_attribute(p_key, p_value: String);

                      function f_display_xml_tag: String;
                      function f_display_xml_name_and_attributes: String;
                      function f_display_attributes: String;
                      function f_display_attribute_value(p_attribute: String): String;
                      function f_display_name_attribute: String;
                      procedure display_strings;
                      procedure display_xml_tag;

                      function f_contains_string: Boolean;
                      function f_key_value(p_key: String): String;
                      function f_c_get_strings_list: tStringList;

                      Destructor Destroy; Override;
                    end; // c_xml_tag

         // -- whatever is between <xxx> and </xxx>
         c_xml_tag_content_list= Class(c_basic_object)
                                   _m_c_xml_tag_content_list: tStringList;

                                   Constructor create_xml_tag_content_list(p_name: String);

                                   function f_xml_tag_content_count: Integer;
                                   function f_c_xml_tag_content(p_xml_tag_content_index: Integer): c_xml_string;
                                   function f_index_of(p_xml_tag_content_name: String): Integer;
                                   procedure add_xml_string(p_c_xml_string: c_xml_string);

                                   function f_c_find_by_xml_tag_content(p_xml_tag_content_name: String): c_xml_tag;
                                   procedure display_xml_tag_content_list;

                                   function f_contains_string: Boolean;
                                   procedure display_strings;
                                   function f_c_get_strings_list: tStringList;

                                   Destructor Destroy; Override;
                                 end; // c_xml_tag_content_list

         t_handle_tag= Procedure(p_c_xml_tag: c_xml_tag) of object;

         c_xml_text= Class(c_basic_object)
                       m_c_xml_comment: c_xml_tag;

                       m_c_xml_content: c_xml_tag;

                       Constructor create_xml_text(p_name: String);
                       procedure display_xml_text;
                       procedure redisplay_xml_text;
                       procedure enumerate(p_tag_name: String; p_c_handle_tag: t_handle_tag);
                       procedure check_xml_strings;

                       function f_c_search_tag_value(p_tag, p_value: String): c_xml_tag;
                       function f_c_search_tag_value_nested(p_tag, p_value: String): c_xml_tag;

                       Destructor Destroy; Override;
                     end; // c_xml_text

  implementation
    uses SysUtils, u_strings, u_c_display;

    // -- c_xml_string

    Constructor c_xml_string.create_xml_string(p_name: String;
        p_c_parent_xml_text: c_xml_text);
      begin
        Inherited create_basic_object(p_name);

        m_c_parent_xml_text:= p_c_parent_xml_text;
      end; // create_xml_string

    function c_xml_string.f_c_self: c_xml_string;
      begin
        Result:= Self;
      end; // f_c_self

    function c_xml_string.f_display_xml_string: String;
      begin
        Result:= m_name
      end; // f_display_xml_string

    // -- c_xml_tag

    Constructor c_xml_tag.create_xml_tag(p_name: String;
        p_c_parent_xml_text: c_xml_text);
      begin
        Inherited create_xml_string(p_name, p_c_parent_xml_text);
        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;

        m_c_xml_tag_content_list:= c_xml_tag_content_list.create_xml_tag_content_list('content');
      end; // create_xml_tag

    function c_xml_tag.f_c_self: c_xml_tag;
      begin
        Result:= Self;
      end; // f_c_self

    procedure c_xml_tag.add_attribute(p_key, p_value: String);
      begin
        m_c_key_list.Add(p_key);
        m_c_value_list.Add(p_value);
      end; // add_attribute

    function c_xml_tag.f_display_xml_tag: String;
      begin
        Result:= m_name;
      end; // f_display_xml_tag

    function c_xml_tag.f_display_attributes: String;
      var l_key_index: Integer;
      begin
        Result:= '';
        with m_c_key_list do
          for l_key_index:= 0 to Count- 1 do
          begin
            Result:= Result+ Strings[l_key_index]
              + '="'+ m_c_value_list[l_key_index]+ '"';
            if l_key_index< Count- 1
              then Result:= Result+ ' ';
          end;
      end; // f_display_attributes

    function c_xml_tag.f_display_xml_name_and_attributes: String;
      var l_key_index: Integer;
      begin
        Result:= f_display_xml_tag;

        with m_c_key_list do
          if Count<> 0
            then begin
                Result:= Result+ '[';
                if m_c_value_list.Count<> Count
                  then display_bug_stop('key_count<> value_count');
                for l_key_index:= 0 to Count- 1 do
                begin
                  Result:= Result+ Strings[l_key_index]
                    + '='+ m_c_value_list[l_key_index];
                  if l_key_index< Count- 1
                    then Result:= Result+ ' ';
                end;
                Result:= Result+ ']';
              end;
      end; // f_display_xml_name_and_attributes

    function c_xml_tag.f_display_attribute_value(p_attribute: String): String;
      var l_attribute_index: Integer;
      begin
        with m_c_key_list do
        begin
          l_attribute_index:= IndexOf(p_attribute);
          if l_attribute_index>= 0
            then Result:= m_c_value_list[l_attribute_index]
            else Result:= '';
        end; // with m_c_key_list
      end; // f_display_attribute_value

    function c_xml_tag.f_display_name_attribute: String;
      var l_name_attribute_index: Integer;
      begin
        Result:= f_display_attribute_value('name');
      end; // f_display_name_attribute

    procedure c_xml_tag.display_strings;
      begin
        m_c_xml_tag_content_list.display_strings;
      end; // display_strings

    procedure c_xml_tag.display_xml_tag;
      begin
        display(f_display_xml_tag);

        // -- the attributes as a list

        // -- the content
        if m_c_xml_tag_content_list.f_xml_tag_content_count> 0
          then m_c_xml_tag_content_list.display_xml_tag_content_list;
      end; // display_xml_tag

    function c_xml_tag.f_contains_string: Boolean;
      begin
        Result:= m_c_xml_tag_content_list.f_contains_string;
      end; // f_contains_string

    function c_xml_tag.f_key_value(p_key: String): String;
      var l_index_of: Integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of>= 0
          then Result:= m_c_value_list[l_index_of]
          else Result:= '';
      end; // f_key_value

    function c_xml_tag.f_c_get_strings_list: tStringList;
      begin
        Result:= m_c_xml_tag_content_list.f_c_get_strings_list;
      end; // f_c_get_strings_list

    Destructor c_xml_tag.Destroy;
      begin
        m_c_xml_tag_content_list.Free;
        m_c_key_list.Free;
        m_c_value_list.Free;

        InHerited;
      end; // Destroy

    // -- c_xml_tag_content_list

    Constructor c_xml_tag_content_list.create_xml_tag_content_list(p_name: String);
      begin
        Inherited create_basic_object(p_name);

        _m_c_xml_tag_content_list:= tStringList.Create;
      end; // create_xml_tag_content_line

    function c_xml_tag_content_list.f_xml_tag_content_count: Integer;
      begin
        Result:= _m_c_xml_tag_content_list.Count;
      end; // f_xml_tag_content_count

    function c_xml_tag_content_list.f_c_xml_tag_content(p_xml_tag_content_index: Integer): c_xml_string;
      begin
        Result:= c_xml_string(_m_c_xml_tag_content_list.Objects[p_xml_tag_content_index]);
      end; //  f_c_xml_tag_content

    function c_xml_tag_content_list.f_index_of(p_xml_tag_content_name: String): Integer;
      begin
        Result:= _m_c_xml_tag_content_list.IndexOf(p_xml_tag_content_name);
      end; // f_index_of

    function c_xml_tag_content_list.f_c_find_by_xml_tag_content(p_xml_tag_content_name: String): c_xml_tag;
      var l_index_of: Integer;
      begin
        l_index_of:= f_index_of(p_xml_tag_content_name);
        if l_index_of< 0
          then Result:= Nil
          else Result:= c_xml_tag(_m_c_xml_tag_content_list.Objects[l_index_of]);
      end; // f_c_find_by_name

    procedure c_xml_tag_content_list.add_xml_string(p_c_xml_string: c_xml_string);
      begin
        if _m_c_xml_tag_content_list= Nil
          then display_bug_halt('no_list '+ m_name);
        if p_c_xml_string= Nil
          then display_bug_halt('no_tag ');
        _m_c_xml_tag_content_list.AddObject(p_c_xml_string.m_name, p_c_xml_string);
      end; // add_xml_string

    procedure c_xml_tag_content_list.display_xml_tag_content_list;
      var l_xml_tag_content_index: Integer;
      begin
        display(m_name+ ' '+ IntToStr(f_xml_tag_content_count));

        for l_xml_tag_content_index:= 0 to f_xml_tag_content_count- 1 do
          display(f_c_xml_tag_content(l_xml_tag_content_index).f_display_xml_string);
      end; // display_xml_tag_content_list

    function c_xml_tag_content_list.f_contains_string: Boolean;
      var l_xml_tag_content_index: Integer;
      begin
        Result:= False;
        for l_xml_tag_content_index:= 0 to f_xml_tag_content_count- 1 do
          if not (f_c_xml_tag_content(l_xml_tag_content_index) is c_xml_tag)
            then begin
                Result:= True;
                Break;
              end;
      end; // f_contains_string

    procedure c_xml_tag_content_list.display_strings;
      var l_content_index: Integer;
      begin
        for l_content_index:= 0 to f_xml_tag_content_count- 1 do
          if not (f_c_xml_tag_content(l_content_index) is c_xml_tag)
            then display('  '+ f_c_xml_tag_content(l_content_index).m_name);
      end; // display_strings

    function c_xml_tag_content_list.f_c_get_strings_list: tStringList;
      var l_content_index: Integer;
      begin
        Result:= tStringList.Create;
        for l_content_index:= 0 to f_xml_tag_content_count- 1 do
          if not (f_c_xml_tag_content(l_content_index) is c_xml_tag)
            then Result.Add(f_c_xml_tag_content(l_content_index).m_name);
      end; // f_c_get_strings_list

    Destructor c_xml_tag_content_list.Destroy;
      var l_xml_tag_content_index: Integer;
      begin
        for l_xml_tag_content_index:= 0 to f_xml_tag_content_count- 1 do
          f_c_xml_tag_content(l_xml_tag_content_index).Free;
        _m_c_xml_tag_content_list.Free;

        Inherited;
      end; // Destroy

    // -- c_xml_text

    Constructor c_xml_text.create_xml_text(p_name: String);
      begin
        Inherited create_basic_object(p_name);
      end; // create_xml_text

    procedure c_xml_text.display_xml_text;

      procedure display_content_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            // display('> '+ m_name); // f_display_xml_name_and_attributes);
            // display(f_display_attributes);
            display('> '+ f_display_xml_name_and_attributes);

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                if f_c_xml_tag_content(l_content_index) is c_xml_tag
                  then display_content_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag)
                  else display(''+ f_c_xml_tag_content(l_content_index).m_name+ '_not_xml_tag');
            change_indentation(-2);
          end; // with p_c_xml_tag
        end; // display_content_recursive

      begin // display_xml_text
        // display(m_c_xml_comment.f_display_xml_tag);

        display_content_recursive(0, m_c_xml_content);
        // m_c_xml_content.display_xml_tag;
      end; // display_xml_text

    procedure c_xml_text.enumerate(p_tag_name: String; p_c_handle_tag: t_handle_tag);

      procedure enumerate_recursive(p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            if m_name= p_tag_name
              then p_c_handle_tag(f_c_self);
            // display('> '+ f_display_xml_name_and_attributes);

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                if f_c_xml_tag_content(l_content_index) is c_xml_tag
                  then enumerate_recursive(f_c_xml_tag_content(l_content_index) as c_xml_tag)
            // change_indentation(-2);
          end; // with p_c_xml_tag
        end; // enumerate_recursive

      begin // enumerate
        enumerate_recursive(m_c_xml_content);
      end; // enumerate

    procedure c_xml_text.check_xml_strings;
        // -- debug: check that strings containt no RETURN (paper bug)

      procedure check_xml_strings_recursive(p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
            l_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                if f_c_xml_tag_content(l_content_index) is c_xml_tag
                  then check_xml_strings_recursive(f_c_xml_tag_content(l_content_index) as c_xml_tag)
                  else begin
                      for l_index:= 1 to Length(m_name) do
                        if m_name[l_index] in [chr(13), chr(10)]
                          then display_bug_halt('non_printable');
                    end;
          end; // with p_c_xml_tag
        end; // check_xml_strings_recursive

      begin // check_xml_strings
        check_xml_strings_recursive(m_c_xml_content);
      end; // check_xml_strings

    procedure c_xml_text.redisplay_xml_text;
        // -- redisplay original indented
        // -- if nothing btw <xxx> and </xxx>, replaces with <xxx/>

      procedure display_content_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
            l_spaces: String;
            l_attributes: String;
            l_begin_tag, l_end_tag: String;
        begin
          l_spaces:= ' '+ f_spaces(p_level* 2);
          with p_c_xml_tag do
          begin
            l_begin_tag:= l_spaces+ '<'+ m_name;

            l_attributes:= f_display_attributes;
            if l_attributes<> ''
              then l_begin_tag:= l_begin_tag+ ' '+ l_attributes;
            // display('> '+ f_display_xml_name_and_attributes);

            l_end_tag:= l_spaces+ '</'+ m_name+ '>';
            with m_c_xml_tag_content_list do
            begin
              if f_xml_tag_content_count> 0
                then begin
                    l_begin_tag:= l_begin_tag+ '>';
                    display(l_begin_tag);
                    // display('+ '''+ l_begin_tag+ '''');

                    for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                      if f_c_xml_tag_content(l_content_index) is c_xml_tag
                        then display_content_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag)
                        else display(''+ f_c_xml_tag_content(l_content_index).m_name+ '');

                    // change_indentation(-2);
                    display(l_end_tag);
                    // display('+ '''+ l_end_tag+ '''');
                  end
                else begin
                    l_begin_tag:= l_begin_tag+ '/>';
                    display(l_begin_tag);
                    // display('+ '''+ l_begin_tag+ '''');
                  end;
            end;
          end; // with p_c_xml_tag
        end; // display_content_recursive

      begin // redisplay_xml_text
        // display(m_c_xml_comment.f_display_xml_tag);

        display_content_recursive(0, m_c_xml_content);
      end; // redisplay_xml_text

    function c_xml_text.f_c_search_tag_value(p_tag, p_value: String): c_xml_tag;

      procedure search_tag_value_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
            l_key_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            l_key_index:= m_c_key_list.IndexOf(p_tag);
            if (l_key_index>= 0) and (m_c_value_list[l_key_index]= p_value)
              then begin
                  Result:= f_c_self;
                  // display('found '+ f_display_xml_name_and_attributes);
                  Exit;
                end;

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
              begin
                if f_c_xml_tag_content(l_content_index) is c_xml_tag
                  then search_tag_value_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag);
                if Result<> Nil
                  then Break;
              end;
          end; // with p_c_xml_tag
        end; // search_tag_value_recursive

      begin // f_c_search_tag_value
        Result:= Nil;
        search_tag_value_recursive(0, m_c_xml_content);
      end; // f_c_search_tag_value

    function c_xml_text.f_c_search_tag_value_nested(p_tag, p_value: String): c_xml_tag;

      procedure search_tag_value_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_content_index: Integer;
            l_key_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            l_key_index:= m_c_key_list.IndexOf(p_tag);
            if (l_key_index>= 0) and (m_c_value_list[l_key_index]= p_value)
              then begin
                  Result:= f_c_self;
                  // display('found '+ f_display_xml_name_and_attributes);
                  Exit;
                end;

            if m_c_xml_nested_text<> Nil
              then begin
                  Result:= m_c_xml_nested_text.f_c_search_tag_value_nested(p_tag, p_value);
                  if Result<> Nil
                    then Exit;
                end;

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
              begin
                if f_c_xml_tag_content(l_content_index) is c_xml_tag
                  then search_tag_value_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag);
                if Result<> Nil
                  then Break;
              end;
          end; // with p_c_xml_tag
        end; // search_tag_value_recursive

      begin // f_c_search_tag_value_nested
        Result:= Nil;
        search_tag_value_recursive(0, m_c_xml_content);
      end; // f_c_search_tag_value_nested

    Destructor c_xml_text.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_xml_tag_content_list
    end. // u_c_xml_tag_content_list
