// 002 u_c_xsd_viewer
// 17 sep 2006

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_xsd_viewer;
  interface
    uses Classes, ComCtrls, u_c_basic_object
      , u_c_xml_scanner_for_xsd
      , u_c_xml_parser_nested
      , u_c_xml_tree_nested
    ;

    type c_xsd_viewer= class(c_basic_object)
                         m_c_xml_text: c_xml_text;

                         Constructor create_xsd_viewer(p_name: String);

                         procedure build_xml_text(p_path, p_recursive_file_name: String);

                         procedure display_text(p_c_result_list: tStringList);

                         procedure find_refs;
                         procedure find_types;
                         procedure display_no_ref_no_types_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
                             p_ref, p_type: Boolean; p_c_result_list: tStringList);
                         procedure display_forward_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
                             p_c_result_list: tStringList);

                         procedure build_treeview(p_c_treeview: tTreeView);
                         procedure build_small_treeview(p_c_treeview: tTreeView);
                         procedure build_iebnf(p_c_result_list: tStrings);

                         Destructor Destroy; Override;
                       end; // c_xsd_viewer

  implementation
    uses SysUtils, u_strings, u_c_display;

    // -- c_xsd_viewer

    Constructor c_xsd_viewer.create_xsd_viewer(p_name: String);
      begin
        Inherited create_basic_object(p_name);
      end; // create_xsd_viewer

    procedure c_xsd_viewer.build_xml_text(p_path, p_recursive_file_name: String);

      procedure build_recursive_text(p_recursive_file_name: String; var pv_c_xml_text: c_xml_text);

        procedure enumerate_recursive(p_c_xml_tag: c_xml_tag);
          var l_content_index: Integer;
              l_include_file_name: String;
          begin
            with p_c_xml_tag do
            begin
              if m_name= 'xs:include'
                then begin
                    display('INCL '+ f_display_xml_name_and_attributes);
                    l_include_file_name:= m_c_value_list[0];
                    build_recursive_text(l_include_file_name, m_c_xml_nested_text);
                  end;

              with m_c_xml_tag_content_list do
                for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                  if f_c_xml_tag_content(l_content_index) is c_xml_tag
                    then enumerate_recursive(f_c_xml_tag_content(l_content_index) as c_xml_tag)
            end; // with p_c_xml_tag
          end; // enumerate_recursive

        var l_c_xml_scanner: c_xml_scanner;

        begin // build_recursive_tree
          display('> build '+ p_recursive_file_name);
          l_c_xml_scanner:= c_xml_scanner.create_xml_scanner('scanner',
              p_path+ p_recursive_file_name);
          pv_c_xml_text:= c_xml_text.create_xml_text('xml_text '+ p_recursive_file_name);

          with c_xml_parser.create_xml_parser('parser', l_c_xml_scanner) do
          begin
            parse_xml(pv_c_xml_text);

            enumerate_recursive(pv_c_xml_text.m_c_xml_content);

            Free;
          end; // with c_xml_parser
          display('< build');
        end; // build_recursive_tree

      begin // build_xml_text
        build_recursive_text(p_recursive_file_name, m_c_xml_text);
      end; // build_xml_text

    procedure c_xsd_viewer.display_text(p_c_result_list: tStringList);

      procedure display_recursive_tree(p_level: Integer; p_c_xml_tag: c_xml_tag);
        var l_document_index: Integer;
            l_content_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            if (m_name= 'xs:annotation')
              then begin
                  with m_c_xml_tag_content_list do
                  begin
                    if (f_xml_tag_content_count> 0)
                        and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                      then begin
                          for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                            with f_c_xml_tag_content(l_document_index) do
                              if m_name= 'xs:documentation'
                                then
                                  p_c_result_list.Add(f_spaces(p_level* 2)
                                    + '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                          Exit;
                        end
                      else display_bug_stop('not 1 annotation');
                  end;
                end;

            p_c_result_list.Add(f_spaces(p_level* 2)+ f_display_xml_name_and_attributes);

            if m_c_xml_nested_text<> Nil
              then begin
                  p_c_result_list.Add('');
                  display_recursive_tree(p_level+ 2, m_c_xml_nested_text.m_c_xml_content);
                  p_c_result_list.Add('');
                end;

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                with f_c_xml_tag_content(l_content_index) do
                  if f_c_self is c_xml_tag
                    then display_recursive_tree(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag)
                    else begin
                        // display(''+ f_c_xml_tag_content(l_content_index).m_name+ '_not_xml_tag');
                      end;
          end; // with p_c_xml_tag
        end; // display_recursive_tree

      begin // display_text
        display_recursive_tree(0, m_c_xml_text.m_c_xml_content);
      end; // display_text

    procedure c_xsd_viewer.find_refs;

      procedure find_refs_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);

        function f_c_search_value(p_value: String): c_xml_tag;
          begin
            Result:= m_c_xml_text.f_c_search_tag_value_nested('name', p_value);
            if Result<> Nil
              then begin
                  display(f_spaces(p_level* 2+ 2)+ 'FOUND '+ Result.f_display_xml_name_and_attributes);
                  Result.m_is_ref:= True;
                end;
          end; // search_value

        var l_ref_index: Integer;
            l_content_index: Integer;

        begin // find_refs_recursive
          with p_c_xml_tag do
          begin
            // -- find all the tags with "ref="
            if m_c_key_list<> Nil
              then begin
                  l_ref_index:= m_c_key_list.IndexOf('ref');
                  if l_ref_index>= 0
                    then begin
                        display(f_spaces(p_level* 2)+ f_display_xml_name_and_attributes);
                        m_c_ref_tag:= f_c_search_value(m_c_value_list[l_ref_index]);
                      end;
                end;

            if m_c_xml_nested_text<> Nil
              then find_refs_recursive(p_level+ 2, m_c_xml_nested_text.m_c_xml_content);

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                with f_c_xml_tag_content(l_content_index) do
                  if f_c_self is c_xml_tag
                    then find_refs_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag);
          end; // with p_c_xml_tag
        end; // find_refs_recursive

      begin // find_refs
        find_refs_recursive(0, m_c_xml_text.m_c_xml_content);
      end; // find_refs

    procedure c_xsd_viewer.find_types;

      procedure find_types_recursive(p_level: Integer; p_c_xml_tag: c_xml_tag);

        function f_c_search_value(p_value: String): c_xml_tag;
          begin
            Result:= m_c_xml_text.f_c_search_tag_value_nested('name', p_value);
            if Result<> Nil
              then begin
                  display(f_spaces(p_level* 2+ 2)+ 'FOUND '+ Result.f_display_xml_name_and_attributes
                      + ' '+ Result.m_c_parent_xml_text.m_name);
                  Result.m_is_type:= True;
                end;
          end; // f_c_search_value

        var l_type_index: Integer;
            l_content_index: Integer;

        begin // find_types_recursive
          with p_c_xml_tag do
          begin
            // -- find all the tags with "type="
            if m_c_key_list<> Nil
              then begin
                  l_type_index:= m_c_key_list.IndexOf('type');
                  if l_type_index>= 0
                    then begin
                        display(f_spaces(p_level* 2)+ f_display_xml_name_and_attributes);
                        m_c_type_tag:= f_c_search_value(m_c_value_list[l_type_index]);
                      end;
                end;

            if m_c_xml_nested_text<> Nil
              then find_types_recursive(p_level+ 2, m_c_xml_nested_text.m_c_xml_content);

            with m_c_xml_tag_content_list do
              for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                with f_c_xml_tag_content(l_content_index) do
                  if f_c_self is c_xml_tag
                    then find_types_recursive(p_level+ 1, f_c_xml_tag_content(l_content_index) as c_xml_tag);
          end; // with p_c_xml_tag
        end; // find_types_recursive

      begin // find_types
        find_types_recursive(0, m_c_xml_text.m_c_xml_content);
      end; // find_types

    procedure c_xsd_viewer.display_no_ref_no_types_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
        p_ref, p_type: Boolean; p_c_result_list: tStringList);

      procedure display_the_tag;
        begin
          with p_c_xml_tag do
          begin
            if (m_c_ref_tag<> Nil) and (m_c_type_tag<> nil)
              then display_bug_stop('both_ref_and_type '+ p_c_xml_tag.m_name);

            if m_c_ref_tag<> nil
              then begin
                  // -- replace the tag display by its ref display
                  display_no_ref_no_types_recursive(p_level, 'ref ', m_c_ref_tag, true, false, p_c_result_list);
                end
              else
                if m_c_type_tag<> Nil
                  then begin
                      p_c_result_list.Add(f_spaces(p_level* 2)+ 'TYP '+ f_display_xml_name_and_attributes);
                      display_no_ref_no_types_recursive(p_level+ 1, 'typ ', m_c_type_tag, false, true, p_c_result_list);
                    end;
          end; // with p_c_xml_tag
        end; // display_the_tag

      var l_document_index: Integer;
          l_content_index: Integer;

      begin // display_no_ref_no_types_recursive
        with p_c_xml_tag do
        begin
          if (not m_is_ref or p_ref) and (not m_is_type or p_type)
            then begin
                if (m_name= 'xs:annotation')
                  then begin
                      with m_c_xml_tag_content_list do
                      begin
                        if (f_xml_tag_content_count> 0)
                            and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                          then begin
                              for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                                with f_c_xml_tag_content(l_document_index) do
                                  if m_name= 'xs:documentation'
                                    then
                                      p_c_result_list.Add(f_spaces(p_level* 2)
                                        + '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                              Exit;
                            end
                          else display_bug_stop('not 1 annotation');
                      end;
                    end;

                display_the_tag;

                with m_c_xml_tag_content_list do
                  for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_content_index) do
                      if f_c_self is c_xml_tag
                        then display_no_ref_no_types_recursive(p_level+ 1, p_title, f_c_xml_tag_content(l_content_index) as c_xml_tag, p_ref, p_type, p_c_result_list)
              end
            else // p_c_result_list.Add(f_spaces(p_level* 2)+ 'ref_typ '+ f_display_xml_name_and_attributes); ;
        end; // with p_c_xml_tag
      end; // display_no_ref_no_types_recursive

    procedure c_xsd_viewer.display_forward_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
        p_c_result_list: tStringList);

      function f_display_annotation(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
        var l_document_index: Integer;
        begin
          Result:= False;
          with p_c_xml_tag do
            if (m_name= 'xs:annotation')
              then begin
                  Result:= True;
                  with m_c_xml_tag_content_list do
                  begin
                    if (f_xml_tag_content_count> 0)
                        and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                      then begin
                          for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                            with f_c_xml_tag_content(l_document_index) do
                              if m_name= 'xs:documentation'
                                then
                                  p_c_result_list.Add(f_spaces(p_level* 2)
                                    + '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                          Exit;
                        end
                      else display_bug_stop('not 1 annotation');
                  end;
              end; // annotation  
        end; // f_display_annotation

      function f_display_simple_type(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
        // xs:simpleType[name=TexteCourt]
        //   xs:restriction[base=Texte]
        //     xs:maxLength[value=35]
        var l_content_index: Integer;
            l_result: String;
        begin
          Result:= False;
          with p_c_xml_tag do
          begin
            if m_name= 'xs:simpleType'
              then begin
                  with m_c_xml_tag_content_list do
                  begin
                    if f_xml_tag_content_count<= 0
                      then display_bug_stop('no_restriction')
                      else
                        for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                          with f_c_xml_tag_content(l_content_index) do
                            if (f_c_self is c_xml_tag) and (m_name= 'xs:restriction')
                              then
                                with f_c_self as c_xml_tag do
                                begin
                                  l_result:= f_display_xml_name_and_attributes;
                                  p_c_result_list.Add(f_spaces(p_level* 2)+ 'simp '+ l_result);
                                  Result:= True;
                                end;
                  end;
                end;
          end; // with p_c_xml_tag
        end; // f_display_simple_type

      procedure display_the_tag_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
          p_display_children: Boolean);
        var l_content_index: Integer;
        begin
          with p_c_xml_tag do
          begin
            if (m_c_ref_tag<> Nil) and (m_c_type_tag<> nil)
              then display_bug_stop('both_ref_and_type '+ p_c_xml_tag.m_name);

            if m_c_ref_tag<> nil
              then begin
                  // -- replace the tag display by its ref display
                  display_the_tag_recursive(p_level, 'ref ', m_c_ref_tag, true);
                end
              else
                if m_c_type_tag<> Nil
                  then begin
                      p_c_result_list.Add(f_spaces(p_level* 2)+ 'TYP '+ f_display_xml_name_and_attributes);
                      display_the_tag_recursive(p_level+ 1, 'typ ', m_c_type_tag, true);

                    end
                  else begin
                      if not f_display_annotation(p_level, f_c_self)
                          and not f_display_simple_type(p_level, f_c_self)
                        then begin
                            p_c_result_list.Add(f_spaces(p_level* 2)+ p_title+ f_display_xml_name_and_attributes);

                            if p_display_children
                              then begin
                                  // -- recurse on the sub elements
                                  with m_c_xml_tag_content_list do
                                    for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                                      with f_c_xml_tag_content(l_content_index) do
                                        if f_c_self is c_xml_tag
                                          then display_the_tag_recursive(p_level+ 1, 'child ', f_c_xml_tag_content(l_content_index) as c_xml_tag, True)
                                end;
                          end;
                    end;
          end; // with p_c_xml_tag
        end; // display_the_tag_recursive

      var l_document_index: Integer;
          l_content_index: Integer;

      begin // display_forward_recursive
        with p_c_xml_tag do
        begin
          if not m_is_ref and not m_is_type
            then begin
                if (m_name= 'xs:annotation')
                  then begin
                      with m_c_xml_tag_content_list do
                      begin
                        if (f_xml_tag_content_count> 0)
                            and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                          then begin
                              for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                                with f_c_xml_tag_content(l_document_index) do
                                  if m_name= 'xs:documentation'
                                    then
                                      p_c_result_list.Add(f_spaces(p_level* 2)
                                        + '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                              Exit;
                            end
                          else display_bug_stop('not 1 annotation');
                      end;
                    end;

                display_the_tag_recursive(p_level, 'MAIN ', p_c_xml_tag, False);

                with m_c_xml_tag_content_list do
                  for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_content_index) do
                      if f_c_self is c_xml_tag
                        then display_forward_recursive(p_level+ 1, p_title, f_c_xml_tag_content(l_content_index) as c_xml_tag, p_c_result_list)
              end;
        end; // with p_c_xml_tag
      end; // display_forward_recursive

    procedure c_xsd_viewer.build_treeview(p_c_treeview: tTreeView);
      var l_c_current_tree_node: TTreeNode;

      procedure add_treeview_node(p_level: Integer; p_text: String);
        var l_c_parent_node: tTreeNode;
        begin
          // -- if first node, just create
          if l_c_current_tree_node = nil
            then l_c_current_tree_node:= p_c_treeview.Items.AddChild(nil, p_text)
            else
              // -- if level did not change, add as a child
              if p_level= l_c_current_tree_node.Level
                then l_c_current_tree_node:= p_c_treeview.Items.AddChild(l_c_current_tree_node.Parent, p_text)
                else
                  // -- if one level below, add below
                  if p_level= l_c_current_tree_node.Level+ 1
                    then l_c_current_tree_node:= p_c_treeview.Items.AddChild(l_c_current_tree_node, p_text)
                    else
                      // -- level changed: find the higher level
                      if p_level< l_c_current_tree_node.Level
                        then begin
                            l_c_parent_node:= l_c_current_tree_node.Parent;
                            while l_c_parent_node.Level > p_level do
                              l_c_parent_node:= l_c_parent_node.Parent;

                            l_c_current_tree_node:= p_c_treeview.Items.AddChild(l_c_parent_node.Parent, p_text);
                          end
                        else display_bug_halt('add_node: leve_error');
        end; // add_treeview_node

      procedure build_treeview_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag);

        function f_display_annotation(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          var l_document_index: Integer;
          begin
            Result:= False;
            with p_c_xml_tag do
              if (m_name= 'xs:annotation')
                then begin
                    Result:= True;
                    with m_c_xml_tag_content_list do
                    begin
                      if (f_xml_tag_content_count> 0)
                          and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                        then begin
                            for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                              with f_c_xml_tag_content(l_document_index) do
                                if m_name= 'xs:documentation'
                                  then
                                    add_treeview_node(p_level, '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                            Exit;
                          end
                        else display_bug_stop('not 1 annotation');
                    end;
                end; // annotation
          end; // f_display_annotation

        function f_display_simple_type(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          // xs:simpleType[name=TexteCourt]
          //   xs:restriction[base=Texte]
          //     xs:maxLength[value=35]
          var l_content_index: Integer;
              l_result: String;
          begin
            Result:= False;
            with p_c_xml_tag do
            begin
              if m_name= 'xs:simpleType'
                then begin
                    with m_c_xml_tag_content_list do
                    begin
                      if f_xml_tag_content_count<= 0
                        then display_bug_stop('no_restriction')
                        else
                          for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                            with f_c_xml_tag_content(l_content_index) do
                              if (f_c_self is c_xml_tag) and (m_name= 'xs:restriction')
                                then
                                  with f_c_self as c_xml_tag do
                                  begin
                                    l_result:= f_display_xml_name_and_attributes;
                                    add_treeview_node(p_level, 'simp '+ l_result);

                                    Result:= True;
                                  end;
                    end;
                  end;
            end; // with p_c_xml_tag
          end; // f_display_simple_type

        procedure display_the_tag_recursive(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
            p_display_children: Boolean);
          var l_content_index: Integer;
          begin
            with p_c_xml_tag do
            begin
              if (m_c_ref_tag<> Nil) and (m_c_type_tag<> nil)
                then display_bug_stop('both_ref_and_type '+ p_c_xml_tag.m_name);

              if m_c_ref_tag<> nil
                then begin
                    // -- replace the tag display by its ref display
                    display_the_tag_recursive(p_level, 'ref ', m_c_ref_tag, true);
                  end
                else
                  if m_c_type_tag<> Nil
                    then begin
                        add_treeview_node(p_level, 'TYP '+ f_display_xml_name_and_attributes);

                        display_the_tag_recursive(p_level+ 1, 'typ ', m_c_type_tag, true);
                      end
                    else begin
                        if not f_display_annotation(p_level, f_c_self)
                            and not f_display_simple_type(p_level, f_c_self)
                          then begin
                              add_treeview_node(p_level, p_title+ f_display_xml_name_and_attributes);

                              if p_display_children
                                then begin
                                    // -- recurse on the sub elements
                                    with m_c_xml_tag_content_list do
                                      for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                                        with f_c_xml_tag_content(l_content_index) do
                                          if f_c_self is c_xml_tag
                                            then display_the_tag_recursive(p_level+ 1, 'child ', f_c_xml_tag_content(l_content_index) as c_xml_tag, True)
                                  end;
                            end;
                      end;
            end; // with p_c_xml_tag
          end; // display_the_tag_recursive

        var l_document_index: Integer;
            l_content_index: Integer;

        begin // build_treeview_recursive
          with p_c_xml_tag do
          begin
            if not m_is_ref and not m_is_type
              then begin
                  if (m_name= 'xs:annotation')
                    then begin
                        with m_c_xml_tag_content_list do
                        begin
                          if (f_xml_tag_content_count> 0)
                              and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                            then begin
                                for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                                  with f_c_xml_tag_content(l_document_index) do
                                    if m_name= 'xs:documentation'
                                      then
                                        add_treeview_node(p_level,
                                           '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                                Exit;
                              end
                            else display_bug_stop('not 1 annotation');
                        end;
                      end;

                  display_the_tag_recursive(p_level, 'MAIN ', p_c_xml_tag, False);

                  with m_c_xml_tag_content_list do
                    for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                      with f_c_xml_tag_content(l_content_index) do
                        if f_c_self is c_xml_tag
                          then build_treeview_recursive(p_level+ 1, p_title, f_c_xml_tag_content(l_content_index) as c_xml_tag)
                end;
          end; // with p_c_xml_tag
        end; // build_treeview_recursive

      begin // build_treeview
        l_c_current_tree_node:= nil;
        p_c_treeview.Items.BeginUpdate;
        build_treeview_recursive(0, 'BAS ', m_c_xml_text.m_c_xml_content);
        p_c_treeview.Items.EndUpdate;
      end; // build_treeview

    procedure c_xsd_viewer.build_small_treeview(p_c_treeview: tTreeView);

      procedure build_node(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
          p_c_parent_treenode: tTreeNode);
          // -- here has only
          // --   - annotations
          // --   - element containing complex types

        function f_build_annotation(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          var l_document_index: Integer;
          begin
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:annotation')
                then begin
                    Result:= True;
                    with m_c_xml_tag_content_list do
                    begin
                      if (f_xml_tag_content_count> 0)
                          and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                        then begin
                            for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                              with f_c_xml_tag_content(l_document_index) do
                                if m_name= 'xs:documentation'
                                  then p_c_treeview.Items.AddChild(p_c_parent_treenode,
                                      '// '+ (f_c_xml_tag_content(l_document_index) as c_xml_tag).m_c_xml_tag_content_list.f_c_xml_tag_content(0).m_name);
                            Exit;
                          end
                        else display_bug_stop('not 1 annotation');
                    end;
                end; // annotation
          end; // f_build_annotation

        function f_c_tag_or_ref(p_c_xml_tag: c_xml_tag): c_xml_tag;
            // -- if the tag is a ref, (recursively) return the ref
          begin
            with p_c_xml_tag do
              if m_c_ref_tag<> Nil
                then Result:= f_c_tag_or_ref(m_c_ref_tag)
                else Result:= p_c_xml_tag
          end; // f_c_tag_or_ref

        function f_build_element(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          var l_c_element_treenode: tTreeNode;
              l_element_content_index: Integer;
          begin // f_build_element
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:element')
                then begin
                    display('> element '+ f_display_xml_name_and_attributes);
                    Result:= True;

                    l_c_element_treenode:= p_c_treeview.Items.AddChild(p_c_parent_treenode, f_display_xml_name_and_attributes);

                    if m_c_type_tag<> Nil
                      then begin
                          display('has_type');
                          // -- unfold the non terminal
                          build_node(p_level+ 1, 'typ '+ p_title,
                              m_c_type_tag,
                              l_c_element_treenode);
                        end
                      else
                        // -- the content of the element
                        with m_c_xml_tag_content_list do
                          for l_element_content_index:= 0 to f_xml_tag_content_count- 1 do
                            with f_c_xml_tag_content(l_element_content_index) do
                              if f_c_self is c_xml_tag
                                then build_node(p_level+ 1, 'child '+ p_title, f_c_xml_tag_content(l_element_content_index) as c_xml_tag,
                                    l_c_element_treenode);
                    display('< element ');
                  end;
          end; // f_build_element

        function f_build_complex_type(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;

          procedure build_choice(p_c_choice_xml_tag: c_xml_tag);
            var l_c_choice_treenode: tTreeNode;
                l_choice_index: Integer;
            begin
              display('> choice');
              with p_c_choice_xml_tag do
              begin
                l_c_choice_treenode:= p_c_treeview.Items.AddChild(p_c_parent_treenode, f_display_xml_name_and_attributes);

                // -- the alternatives
                with m_c_xml_tag_content_list do
                  for l_choice_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_choice_index) do
                      if f_c_self is c_xml_tag
                        then build_node(p_level+ 1, 'child '+ p_title, f_c_xml_tag_content(l_choice_index) as c_xml_tag,
                            l_c_choice_treenode);
              end; // with p_c_xml_tag
              display('< choice');
            end; // build_choice

          procedure build_sequence(p_c_sequence_xml_tag: c_xml_tag);
            var l_c_sequence_treenode: tTreeNode;
                l_sequence_index: Integer;
            begin
              display('> sequence');
              with p_c_sequence_xml_tag do
              begin
                l_c_sequence_treenode:= p_c_treeview.Items.AddChild(p_c_parent_treenode, f_display_xml_name_and_attributes);

                // -- the alternatives
                with m_c_xml_tag_content_list do
                  for l_sequence_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_sequence_index) do
                      if f_c_self is c_xml_tag
                        then build_node(p_level+ 1, 'child '+ p_title, f_c_xml_tag_content(l_sequence_index) as c_xml_tag,
                            l_c_sequence_treenode);
              end; // with p_c_xml_tag
              display('< sequence');
            end; // build_sequence

          var l_complex_index: Integer;

          begin // f_build_complex_type
            display('> complex_type');
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:complexType')
                then begin
                    Result:= True;

                    // -- is it a CHOICE or a SEQUENCE
                    // -- but can have annotations
                    with m_c_xml_tag_content_list do
                      for l_complex_index:= 0 to f_xml_tag_content_count- 1 do
                        with f_c_xml_tag_content(l_complex_index) do
                          if f_c_self is c_xml_tag
                            then
                              if m_name= 'xs:choice'
                                then build_choice(c_xml_tag(f_c_self))
                                else
                                  if m_name= 'xs:sequence'
                                    then build_sequence(c_xml_tag(f_c_self))
                                    else
                                      if m_name= 'xs:annotation'
                                        then f_build_annotation(p_level+ 1, c_xml_tag(f_c_self));
                  end;
            display('< complex_type');
          end; // f_build_complex_type

        begin // build_node
          display('> build_node');
          with p_c_xml_tag do
          begin
            display(m_name);
            if    f_build_annotation(p_level, f_c_self)
                or
                  f_build_element(p_level, f_c_tag_or_ref(f_c_self))
                or
                  f_build_complex_type(p_level, f_c_self)
              then ;
          end; // with p_c_xml_tag
          display('< build_node');
        end; // build_node

      var l_c_root_tree_node: tTreeNode;
          l_root_content_index: Integer;

      begin // build_small_treeview
        clear_display;
        p_c_treeview.Items.BeginUpdate;

        // -- built the root node
        l_c_root_tree_node:= p_c_treeview.Items.AddChild(nil, 'ROOT '+ m_c_xml_text.m_c_xml_content.f_display_xml_name_and_attributes);

        // -- skip the include, until hits the FIRST main (assumes INCLUDE only at the top)
        with m_c_xml_text.m_c_xml_content.m_c_xml_tag_content_list do
          for l_root_content_index:= 0 to f_xml_tag_content_count- 1 do
            with f_c_xml_tag_content(l_root_content_index) do
              if f_c_self is c_xml_tag
                then
                  with c_xml_tag(f_c_xml_tag_content(l_root_content_index)) do
                    if m_name<> 'xs:include'
                      then begin
                          build_node(0, 'BAS ', c_xml_tag(f_c_xml_tag_content(l_root_content_index)), l_c_root_tree_node);
                          Break;
                        end;

        p_c_treeview.Items.EndUpdate;
      end; // build_small_treeview

    procedure c_xsd_viewer.build_iebnf(p_c_result_list: tStrings);

      procedure add_to_result(p_stringlist_index: Integer; p_text: string);
        begin
          if p_stringlist_index<> -1
            then p_c_result_list[p_stringlist_index]:= p_c_result_list[p_stringlist_index]+ p_text;
          display('ADD '+ IntToStr(p_stringlist_index)+ ' '+ p_text);
        end; // add_to_result

      procedure remove_ending_alterative(p_stringlist_index: Integer);
        begin
          p_c_result_list[p_stringlist_index]:= f_remove_end_if_end_is_equal_to(p_c_result_list[p_stringlist_index], '| ');
        end; // remove_ending_alterative

      procedure build_node(p_level: Integer; p_title: String; p_c_xml_tag: c_xml_tag;
          p_stringlist_index: Integer);

        function f_build_annotation(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          var l_document_index: Integer;
          begin
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:annotation')
                then begin
                    Result:= True;
                    with m_c_xml_tag_content_list do
                    begin
                      if (f_xml_tag_content_count> 0)
                          and (f_c_xml_tag_content(0).m_name= 'xs:documentation')
                        then begin
                            for l_document_index:= 0 to f_xml_tag_content_count- 1 do
                              with f_c_xml_tag_content(l_document_index) do
                                if m_name= 'xs:documentation'
                                  then  // could add a comment in the IEBNF
                                        ;
                            Exit;
                          end
                        else display_bug_stop('not 1 annotation');
                    end;
                end; // annotation
          end; // f_build_annotation

        function f_c_tag_or_ref(p_c_xml_tag: c_xml_tag): c_xml_tag;
            // -- if the tag is a ref, (recursively) return the ref
          begin
            with p_c_xml_tag do
              if m_c_ref_tag<> Nil
                then Result:= f_c_tag_or_ref(m_c_ref_tag)
                else Result:= p_c_xml_tag
          end; // f_c_tag_or_ref

        procedure display_repetition_start(p_c_xml_tag: c_xml_tag);
          begin
            with p_c_xml_tag do
            begin
              if f_display_attribute_value('minOccurs')= '0'
                then add_to_result(p_stringlist_index, ' [ ');
              if f_display_attribute_value('maxOccurs')= 'unbounded'
                then add_to_result(p_stringlist_index, ' { ');
            end;
          end; // display_repetition_start

        procedure display_repetition_end(p_c_xml_tag: c_xml_tag);
          begin
            with p_c_xml_tag do
            begin
              if f_display_attribute_value('minOccurs')= '0'
                then add_to_result(p_stringlist_index, ' ] ');
              if f_display_attribute_value('maxOccurs')= 'unbounded'
                then add_to_result(p_stringlist_index, ' } ');
            end;
          end; // display_repetition_end

        var l_debug: String;

        function f_build_simple_type(p_stringlist_index: Integer; p_c_xml_tag: c_xml_tag): Boolean;
          // xs:simpleType[name=TexteCourt]
          //   xs:restriction[base=Texte]
          //     xs:maxLength[value=35]
          var l_content_index: Integer;
              l_result: String;
          begin
            Result:= False;

            with p_c_xml_tag do
            begin
              if m_name= 'xs:simpleType'
                then begin
                    with m_c_xml_tag_content_list do
                    begin
                      if f_xml_tag_content_count<= 0
                        then display_bug_stop('no_restriction')
                        else
                          for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                            with f_c_xml_tag_content(l_content_index) do
                              if (f_c_self is c_xml_tag) and (m_name= 'xs:restriction')
                                then
                                  with f_c_self as c_xml_tag do
                                  begin
                                    l_result:= f_display_attribute_value('base');
                                    if Copy(l_result, 1, 3)= 'xs:'
                                      then Delete(l_result, 1, 3);
                                    if Trim(l_result)= ''
                                      then display_bug_stop('no_base');
                                    l_debug:= l_result+ ' '+ IntToStr(p_stringlist_index);
                                    add_to_result(p_stringlist_index, '"'+ l_result+ '" ');

                                    Result:= True;
                                  end;
                    end;
                  end;
            end; // with p_c_xml_tag
          end; // f_build_simple_type

        function f_build_element(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;

          procedure handle_type;
            var l_non_terminal_index: Integer;
                l_title: String;
            begin
              with p_c_xml_tag do
              begin
                display('has_type');
                l_title:= '';

                // -- prepare the new line for the new non-terminal
                l_non_terminal_index:= p_c_result_list.Add(f_spaces(p_level)+ l_title+ f_display_name_attribute+ '= ');
                // -- unfold the non terminal
                build_node(p_level+ 1, 'typ'+ p_title, m_c_type_tag, l_non_terminal_index);
                add_to_result(l_non_terminal_index, '.');
              end; // with p_c_xml_tag
            end; // handle_type

          procedure handle_ref;

            function f_simple_type_tag_index(p_c_xml_tag: c_xml_tag): Integer;
                // -- either an xs:simpleType or comment+ xs_simpleType
              var l_content_index: Integer;
              begin
                Result:= -1;

                with p_c_xml_tag.m_c_xml_tag_content_list do
                  for l_content_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_content_index) do
                      if m_name= 'xs:annotation'
                        then
                        else
                          if m_name= 'xs:simpleType'
                            then begin
                                Result:= l_content_index;
                              end;
              end; // f_simple_type_tag_index

            var l_simple_type_index: Integer;
                l_c_tag_ref: c_xml_tag;
                l_simple_type_tag_index: Integer;

            begin // handle_ref
              with p_c_xml_tag do
              begin
                display('has_ref');
                // -- if the ref contains a simple type, display the ref name, and add a simple type prod
                add_to_result(p_stringlist_index, f_c_tag_or_ref(f_c_self).f_display_name_attribute+ ' ');

                l_c_tag_ref:= f_c_tag_or_ref(f_c_self);

                with l_c_tag_ref, m_c_xml_tag_content_list do
                begin
                  l_simple_type_tag_index:= f_simple_type_tag_index(l_c_tag_ref);

                  if l_simple_type_tag_index>= 0
                    then begin
                        l_simple_type_index:= p_c_result_list.Add(f_spaces(p_level)+ f_c_tag_or_ref(f_c_self).f_display_name_attribute+ '= ');

                        // -- directly add the simple type
                        if not f_build_simple_type(l_simple_type_index,
                            l_c_tag_ref.m_c_xml_tag_content_list.f_c_xml_tag_content(l_simple_type_tag_index) as c_xml_tag)
                          then ; // display_bug_stop('no_simple');
                        add_to_result(l_simple_type_index, '.');
                      end
                    else begin
                        // -- the ref is not an "xs:simpleType"
                        // -- unfold the non terminal
                        // -- get the children but do not display this level
                        build_node(p_level, 'tc'+ p_title, f_c_tag_or_ref(f_c_self), -1);
                      end;
                end;
              end; // with p_c_xml_tag
            end; // handle_ref

          procedure handle_children;
              // -- the content of the element
            var l_element_content_index: Integer;
            begin
              with p_c_xml_tag, m_c_xml_tag_content_list do
                for l_element_content_index:= 0 to f_xml_tag_content_count- 1 do
                  with f_c_xml_tag_content(l_element_content_index) do
                    if f_c_self is c_xml_tag
                      then build_node(p_level+ 1, 'child '+ p_title,
                          f_c_xml_tag_content(l_element_content_index) as c_xml_tag,
                          p_stringlist_index);
            end; // handle_children

          var l_title: String;

          begin // f_build_element
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:element')
                then begin
                    display('> element '+ f_display_xml_name_and_attributes);
                    Result:= True;
                    // l_title:= 'el_'+ p_title;
                    l_title:= '';

                    if m_c_ref_tag= Nil
                      then begin
                          // -- add the item to the current production
                          display_repetition_start(p_c_xml_tag);
                          add_to_result(p_stringlist_index, l_title+ f_display_name_attribute+ ' ');
                          display_repetition_end(p_c_xml_tag);

                          if m_c_type_tag<> Nil
                            then handle_type
                            else handle_children;
                        end
                      else handle_ref;

                    display('< element ');
                  end;
          end; // f_build_element

        function f_build_complex_type(p_level: Integer; p_c_xml_tag: c_xml_tag): Boolean;

          procedure build_choice(p_c_choice_xml_tag: c_xml_tag);
            var l_choice_index: Integer;
            begin
              display('> choice');
              with p_c_choice_xml_tag do
              begin
                display_repetition_start(p_c_xml_tag);
                add_to_result(p_stringlist_index, '( ');

                // -- the alternatives
                with m_c_xml_tag_content_list do
                  for l_choice_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_choice_index) do
                      if f_c_self is c_xml_tag
                        then begin
                            build_node(p_level+ 1, 'child '+ p_title,
                                f_c_xml_tag_content(l_choice_index) as c_xml_tag,
                                p_stringlist_index);

                            add_to_result(p_stringlist_index, ' | ');
                            // l_c_choice_treenode);
                          end;

                // -- remove the last "|"
                remove_ending_alterative(p_stringlist_index);
                add_to_result(p_stringlist_index, ' ) ');
                display_repetition_end(p_c_xml_tag);
             end; // with p_c_xml_tag
              display('< choice');
            end; // build_choice

          procedure build_sequence(p_c_sequence_xml_tag: c_xml_tag);
            var l_sequence_index: Integer;
            begin
              display('> sequence');
              with p_c_sequence_xml_tag do
              begin
                display_repetition_start(p_c_xml_tag);

                // -- the alternatives
                with m_c_xml_tag_content_list do
                  for l_sequence_index:= 0 to f_xml_tag_content_count- 1 do
                    with f_c_xml_tag_content(l_sequence_index) do
                      if f_c_self is c_xml_tag
                        then build_node(p_level+ 1, 'child '+ p_title,
                            f_c_xml_tag_content(l_sequence_index) as c_xml_tag,
                            p_stringlist_index);
                display_repetition_end(p_c_xml_tag);
              end; // with p_c_xml_tag
              display('< sequence');
            end; // build_sequence

          var l_complex_index: Integer;

          begin // f_build_complex_type
            display('> complex_type');
            Result:= False;

            with p_c_xml_tag do
              if (m_name= 'xs:complexType')
                then begin
                    Result:= True;

                    // -- is it a CHOICE or a SEQUENCE
                    // -- but can have annotations
                    with m_c_xml_tag_content_list do
                      for l_complex_index:= 0 to f_xml_tag_content_count- 1 do
                        with f_c_xml_tag_content(l_complex_index) do
                          if f_c_self is c_xml_tag
                            then
                              if m_name= 'xs:choice'
                                then build_choice(c_xml_tag(f_c_self))
                                else
                                  if m_name= 'xs:sequence'
                                    then build_sequence(c_xml_tag(f_c_self))
                                    else
                                      if m_name= 'xs:annotation'
                                        then f_build_annotation(p_level+ 1, c_xml_tag(f_c_self));
                  end;
            display('< complex_type');
          end; // f_build_complex_type

        begin // build_node
          display('> build_node');
          with p_c_xml_tag do
          begin
            display(m_name);
            if    f_build_annotation(p_level, f_c_self)
                or
                  f_build_simple_type(p_stringlist_index, f_c_self)
                or
                  f_build_element(p_level, f_c_self)
                or
                  f_build_complex_type(p_level, f_c_self)
              then ;
          end; // with p_c_xml_tag
          display('< build_node');
        end; // build_node

      procedure remove_xxx;
        var l_index: Integer;
            l_string: String;
        begin
          l_index:= 0;
          with p_c_result_list do
            while l_index< Count do
            begin
              l_string:= Trim(Strings[l_index]);
              if Pos('xxx', l_string)= 1
                then begin
                    if Pos(' ', l_string)> 0
                      then begin
                          System.Delete(l_string, 1, 3);
                          Strings[l_index]:= l_string;
                        end
                      else Delete(l_index)
                  end
                else Inc(l_index);
            end;
        end; // remove_xxx

      var l_root_content_index: Integer;
          l_iebnf_root_index: Integer;

      begin // build_iebnf
        clear_display;

        // -- skip the include, until hits the FIRST main (assumes INCLUDE only at the top)
        with m_c_xml_text.m_c_xml_content.m_c_xml_tag_content_list do
          for l_root_content_index:= 0 to f_xml_tag_content_count- 1 do
            with f_c_xml_tag_content(l_root_content_index) do
              if f_c_self is c_xml_tag
                then
                  with c_xml_tag(f_c_xml_tag_content(l_root_content_index)) do
                    if m_name<> 'xs:include'
                      then begin
                          l_iebnf_root_index:= p_c_result_list.Add('root ');
                          build_node(0, 'BAS ', c_xml_tag(f_c_xml_tag_content(l_root_content_index)),
                              l_iebnf_root_index);
                          Break;
                        end;
        // remove_xxx;
      end; // build_iebnf

    Destructor c_xsd_viewer.Destroy;
      begin
        // -- todo: free everybody
        Inherited;
      end; // Destroy

    begin // u_c_xsd_viewer
    end. // u_c_xsd_viewer




